;;; $Id: combinators-fix-point.scm,v 1.2 2006/01/05 22:24:09 leavens Exp $
;;; Some fix-point combinators.

(module combinators-fix-point (lib "typedscm.ss" "typedscm")

(provide Y-classical theta Y )

;;; The classical (call by name, lazy) fix point combinator.
;;; This won't work in an eager language like Scheme.
(deftype Y-classical
  (forall (T) (-> ((-> ((-> (T) T)) (-> (T) T))) (-> (T) T))))
(define Y-classical
  (has-type-trusted
    (forall (T) (-> ((-> ((-> (T) T)) (-> (T) T))) (-> (T) T)))
    (lambda (f)
      ((lambda (x) (f (x x)))
       (lambda (x) (f (x x)))))))

;;; Turing's fix-point combinator.
(deftype theta (forall (T) (-> ((-> ((-> (T) T)) (-> (T) T))) (-> (T) T))))
(define theta
  (has-type-trusted
    (forall (T) (-> ((-> ((-> (T) T)) (-> (T) T))) (-> (T) T)))
    (let ((A (lambda (x) (lambda (y) ((y ((x x) y)))))))
      (A A))))

;;; The following is the call-by-value fix point combinator
;;; (Little Lisper p. 156).
;;; This actually works in Scheme.

(deftype Y (forall (T) (-> ((-> ((-> (T) T)) (-> (T) T))) (-> (T) T))))
(define Y
  (has-type-trusted
    (forall (T) (-> ((-> ((-> (T) T)) (-> (T) T))) (-> (T) T)))
    (lambda (f)
      ((lambda (x)
         (f (lambda (arg) ((x x) arg))))
       (lambda (x)
         (f (lambda (arg) ((x x) arg))))))))

;;; See the file fix-mod.scm for a more useful fix-point procedure for Scheme.

;;; Exercise: generalize Y above so that it also works for multiple arguments.

) ;; end module
