;;; $Id: sxml-helpers.tst,v 1.2 2006/01/05 22:24:09 leavens Exp $
;;; AUTHOR: Gary T. Leavens

(require (lib "sxml-helpers.scm" "lib342"))

(define sampledoc0
  '(*TOP*
       (urn:loc.gov:books:book
         (urn:loc.gov:books:title
             "Cheaper by the Dozen")
         (urn:ISBN:0-395-36341-6:number "1568491379")
         (urn:loc.gov:books:notes
           (urn:w3-org-ns:HTML:p
              "This is a "
               (urn:w3-org-ns:HTML:i "funny")
              " book!")))))

(define sampledoc1
  '(*TOP*
       (@ (*NAMESPACES* 
            (HTML "http://www.w3.org/TR/REC-html40")))
       (RESERVATION
         (NAME (@ (HTML:CLASS "largeSansSerif"))
           "Layman, A")
         (SEAT (@ (HTML:CLASS "largeMonotype")
                  (CLASS "Y"))
            "33B")
         (HTML:A (@ (HREF "/cgi-bin/ResStatus"))
            "Check Status")
         (DEPARTURE "1997-05-24T07:55:00+1"))))

(define sample-element0
  '(WEIGHT (@ (unit "pound"))
       (NET (@ (certified)) "67")
       (GROSS "95")
     ))

(define sample-element1
  '(poem (@ (title "The Lovesong of J. Alfred Prufrock")
            (poet "T. S. Eliot"))
         (stanza
          (line "Let us go then, you and I,")
          (line "When the evening is spread out against the sky")
          (line "Like a patient etherized upon a table:"))
         (stanza
          (line "In the room the women come and go")
          (line "Talking of Michaelangelo."))))

(define sample-element2
  '(html
    (head (title "Slides"))
    (body
     (p (@ (align "center"))
        (table (@ (style "font-size: x-large"))
               (tr
                (td (@ (align "right")) "Talks ")
                (td (@ (align "center")) " = ")
                (td " slides + transition"))
               (tr (td)
                   (td (@ (align "center")) " = ")
                   (td " data + control"))
               (tr (td)
                   (td (@ (align "center")) " = ")
                   (td " programs"))))
     (ul
      (li (a (@ (href "slides/slide0001.gif")) "Introduction"))
      (li (a (@ (href "slides/slide0010.gif")) "Summary")))
     )))

(run-regression-tests-equal?
 `(

   ( (document? (document '() '() (named 'html '()))) ==> #t )
   ( (document? sampledoc0) ==> #t )
   ( (document? sampledoc1) ==> #t )
   ( (document? 'x) ==> #f )
   ( (document? 32) ==> #f )
   ( (document? "a val") ==> #f )

   ( (element? (named 'html '())) ==> #t )
   ( (element? (named 'html (list (child (named 'head '()))))) ==> #t )
   ( (element? (named 'html
                      (list (child (named 'head '()))
                            (child (named 'body '()))))) ==> #t )
   ( (element? (attributed 'foo (list (name-value 'temp "32")) '())) ==> #t )
   ( (element? (attributed 'foo
                           (list (name-value 'temp "32"))
                           (list (child (named 'head '()))))) ==> #t )
   ( (element? (attributed 'foo
                           (list (name-value 'temp "32")
                                 (name-value 'clouds "none"))
                           (list (child (named 'head '()))))) ==> #t )
   ( (element? (attributed 'foo
                           (list (name-value 'temp "32")
                                 (name-value 'clouds "none"))
                           (list ))) ==> #t )
   ( (element? (attributed 'foo
                           (list )
                           (list ))) ==> #t )
   ( (element? (attributed 'foo
                           (list )
                           (list (child (named 'head '()))
                                 (child (named 'body '()))))) ==> #t )
   ( (element? (element
                   'report
                 (list (name-value 'temp "32")
                       (name-value 'clouds "none"))
                 (list (child (named 'head '()))
                       (child (named 'body '()))))) ==> #t )
   ( (element? (element
                   'report
                 (list )
                 (list (child (named 'head '()))
                       (child (named 'body '()))))) ==> #t )
   ( (element? sample-element0) ==> #t )
   ( (element? sample-element1) ==> #t )
   ( (element? sample-element2) ==> #t )
   ( (element? 'x) ==> #f )
   ( (element? 32) ==> #f )
   ( (element? "a val") ==> #f )
   ( (element? (comment "a comment")) ==> #f )
   ( (element? (PI 'targ "instr")) ==> #f )

   ( (attribute? (name-value 'phone "555-1212")) ==> #t )
   ( (attribute? (name-only 'warm)) ==> #t )
   ( (attribute? 'warm) ==> #f )
   ( (attribute? "a str") ==> #f )

   ( (child-of-element? (child (named 'html '()))) ==> #t )
   ( (child-of-element? (child sample-element0)) ==> #t )
   ( (child-of-element? (text "a str")) ==> #t )
   ( (child-of-element? (child-pi (PI 'targ "a str"))) ==> #t )
   ( (child-of-element? (child-comment (comment "a str"))) ==> #t )
   ( (child-of-element? (entity "xyz" "adfadfl")) ==> #t )
   ( (child-of-element? 32) ==> #f )

   ( (PI? (PI 'targ "a str")) ==> #t )
   ( (PI? "a str") ==> #f )

   ( (comment? (comment "a str")) ==> #t )
   ( (comment? "a str") ==> #f )

   ( (named? (named 'html '())) ==> #t )
   ( (named? (named 'html (list (child (named 'head '()))))) ==> #t )
   ( (named? (named 'html
                      (list (child (named 'head '()))
                            (child (named 'body '()))))) ==> #t )

   ( (named? (attributed 'foo (list (name-value 'temp "32")) '())) ==> #f )
   ( (named? (attributed 'foo
                           (list (name-value 'temp "32"))
                           (list (child (named 'head '()))))) ==> #f )
   ( (named? (attributed 'foo
                           (list (name-value 'temp "32")
                                 (name-value 'clouds "none"))
                           (list (child (named 'head '()))))) ==> #f )
   ( (named? (attributed 'foo
                           (list (name-value 'temp "32")
                                 (name-value 'clouds "none"))
                           (list ))) ==> #f )
   ( (named? (attributed 'foo
                           (list )
                           (list ))) ==> #f )
   ( (named? (attributed 'foo
                           (list )
                           (list (child (named 'head '()))
                                 (child (named 'body '()))))) ==> #f )

   ( (attributed? (named 'html '())) ==> #f )
   ( (attributed? (named 'html (list (child (named 'head '()))))) ==> #f )
   ( (attributed? (named 'html
                      (list (child (named 'head '()))
                            (child (named 'body '()))))) ==> #f )

   ( (attributed? (attributed 'foo (list (name-value 'temp "32")) '())) ==> #t )
   ( (attributed? (attributed 'foo
                           (list (name-value 'temp "32"))
                           (list (child (named 'head '()))))) ==> #t )
   ( (attributed? (attributed 'foo
                           (list (name-value 'temp "32")
                                 (name-value 'clouds "none"))
                           (list (child (named 'head '()))))) ==> #t )
   ( (attributed? (attributed 'foo
                           (list (name-value 'temp "32")
                                 (name-value 'clouds "none"))
                           (list ))) ==> #t )
   ( (attributed? (attributed 'foo
                           (list )
                           (list ))) ==> #t )
   ( (attributed? (attributed 'foo
                           (list )
                           (list (child (named 'head '()))
                                 (child (named 'body '()))))) ==> #t )

   ( (name-value? (name-value 'temp "32")) ==> #t )
   ( (name-value? (name-only 'cool)) ==> #f )
   ( (name-only? (name-value 'temp "32")) ==> #f )
   ( (name-only? (name-only 'cool)) ==> #t )

   ( (child? (child (name-value 'temp "32"))) ==> #t )
   ( (child? (text "a val")) ==> #f )
   ( (child? (child-comment (comment "a comment"))) ==> #f )
   ( (child? (child-pi (PI 'targ "instr"))) ==> #f )
   ( (child? (entity "abc" "abc-1234")) ==> #f )

   ( (text? (child (name-value 'temp "32"))) ==> #f )
   ( (text? (text "a val")) ==> #t )
   ( (text? (child-comment (comment "a comment"))) ==> #f )
   ( (text? (child-pi (PI 'targ "instr"))) ==> #f )
   ( (text? (entity "abc" "abc-1234")) ==> #f )

   ( (child-pi? (child (name-value 'temp "32"))) ==> #f )
   ( (child-pi? (text "a val")) ==> #f )
   ( (child-pi? (child-comment (comment "a comment"))) ==> #f )
   ( (child-pi? (child-pi (PI 'targ "instr"))) ==> #t )
   ( (child-pi? (entity "abc" "abc-1234")) ==> #f )

   ( (child-comment? (child (name-value 'temp "32"))) ==> #f )
   ( (child-comment? (text "a val")) ==> #f )
   ( (child-comment? (child-comment (comment "a comment"))) ==> #t )
   ( (child-comment? (child-pi (PI 'targ "instr"))) ==> #f )
   ( (child-comment? (entity "abc" "abc-1234")) ==> #f )

   ( (entity? (child (name-value 'temp "32"))) ==> #f )
   ( (entity? (text "a val")) ==> #f )
   ( (entity? (child-comment (comment "a comment"))) ==> #f )
   ( (entity? (child-pi (PI 'targ "instr"))) ==> #f )
   ( (entity? (entity "abc" "abc-1234")) ==> #t )

   ( (document->PIs sampledoc0) ==> () )
   ( (document->PIs sampledoc1) ==> () )
   ( (PI? (PI 'targ "instr")) ==> #t )
   ( (document->PIs (document (list (PI 'targ "instr")) '() (named 'html '())))
     ==> ((*PI* targ "instr")) )
   ( (document->PIs (document (list (PI 'targ "instr")
                                         (PI 'targ2 "instr2"))
                                   (list (comment "a comment")
                                         (comment "another"))
                                   (named 'html '())))
     ==> ((*PI* targ "instr") (*PI* targ2 "instr2")) )

   ( (document->comments sampledoc0) ==> () )
   ( (document->comments sampledoc1) ==> () )
   ( (comment? (comment "a comment")) ==> #t )
   ( (document->comments (document (list (PI 'targ "instr"))
                                   (list (comment "a comment"))
                                   (named 'html '())))
     ==> ((*COMMENT* "a comment")) )

   ( (document->comments (document (list (PI 'targ "instr")
                                         (PI 'targ2 "instr2"))
                                   (list (comment "a comment")
                                         (comment "another"))
                                   (named 'html '())))
     ==> ((*COMMENT* "a comment") (*COMMENT* "another")) )

   ( (document->element sampledoc0)
     ==> (urn:loc.gov:books:book
         (urn:loc.gov:books:title
             "Cheaper by the Dozen")
         (urn:ISBN:0-395-36341-6:number "1568491379")
         (urn:loc.gov:books:notes
           (urn:w3-org-ns:HTML:p
              "This is a "
               (urn:w3-org-ns:HTML:i "funny")
              " book!"))) )
   ( (document->element sampledoc1)
     ==> (RESERVATION
         (NAME (@ (HTML:CLASS "largeSansSerif"))
           "Layman, A")
         (SEAT (@ (HTML:CLASS "largeMonotype")
                  (CLASS "Y"))
            "33B")
         (HTML:A (@ (HREF "/cgi-bin/ResStatus"))
            "Check Status")
         (DEPARTURE "1997-05-24T07:55:00+1")) )
   ( (element? (named 'html '())) ==> #t )
   ( (document->element (document (list (PI 'targ "instr"))
                                   (list (comment "a comment"))
                                   (named 'html '())))
     ==> (html) )

   ( (document->element (document (list (PI 'targ "instr")
                                         (PI 'targ2 "instr2"))
                                   (list (comment "a comment")
                                         (comment "another"))
                                   (named 'html '())))
     ==> (html) )

   ( (element->name (named 'html '()))
     ==> html )
   ( (named->name (named 'html '()))
     ==> html )
   ( (element->name (document->element sampledoc0))
     ==> urn:loc.gov:books:book )
   ( (named->name (document->element sampledoc0))
     ==> urn:loc.gov:books:book )
   ( (element->name (document->element sampledoc1))
     ==> RESERVATION )
   ( (named->name (document->element sampledoc1))
     ==> RESERVATION )
   ( (element->name sample-element0)
     ==> WEIGHT )
   ( (attributed->name sample-element0)
     ==> WEIGHT )
   ( (element->name sample-element1)
     ==> poem )
   ( (attributed->name sample-element1)
     ==> poem )
   ( (element->name sample-element2)
     ==> html )
   ( (named->name sample-element2)
     ==> html )
   ( (named->name (named 'foo '()))
     ==> foo )
   ( (attributed->name (attributed 'html (list (name-value 'temp "34")) '()))
     ==> html )

   ( (named->children (named 'html '()))
     ==> () )
   ( (named->children sample-element2)
     ==> ((head (title "Slides"))
          (body
           (p (@ (align "center"))
              (table (@ (style "font-size: x-large"))
                     (tr
                      (td (@ (align "right")) "Talks ")
                      (td (@ (align "center")) " = ")
                      (td " slides + transition"))
                     (tr (td)
                         (td (@ (align "center")) " = ")
                         (td " data + control"))
                     (tr (td)
                         (td (@ (align "center")) " = ")
                         (td " programs"))))
           (ul
            (li (a (@ (href "slides/slide0001.gif")) "Introduction"))
            (li (a (@ (href "slides/slide0010.gif")) "Summary")))
           )) )

   ( (attributed->children sample-element0)
     ==> ((NET (@ (certified)) "67")
          (GROSS "95")) )

   ( (attributed->children sample-element1)
     ==> ((stanza
           (line "Let us go then, you and I,")
           (line "When the evening is spread out against the sky")
           (line "Like a patient etherized upon a table:"))
          (stanza
           (line "In the room the women come and go")
           (line "Talking of Michaelangelo."))) )

   ( (element->children sample-element2)
     =  (named->children sample-element2) )
   ( (element->children sample-element0)
     =  (attributed->children sample-element0) )
   ( (element->children sample-element1)
     =  (attributed->children sample-element1) )

   ( (attributed->attributes (attributed 'html (list (name-value 'temp "34")) '()))
     ==> ((temp "34")) )
   ( (attributed->attributes sample-element0)
     ==> ((unit "pound")) )

   ( (attributed->attributes sample-element1)
     ==> ((title "The Lovesong of J. Alfred Prufrock")
            (poet "T. S. Eliot")) )

   ( (name-value->name (name-value 'foo "bar")) ==> foo )
   ( (name-value->value (name-value 'foo "bar")) ==> "bar" )
   ( (name-value->name (car (attributed->attributes sample-element0)))
                        ==> unit )
   ( (name-value->value (car (attributed->attributes sample-element0)))
                        ==> "pound" )

   ( (name-only->name (name-only 'foobar)) ==> foobar )

   ( (attribute->name (name-only 'foobar)) ==> foobar )
   ( (attribute->value (name-only 'foobar)) ==> "" )
   ( (attribute->name (name-value 'foo "bar")) ==> foo )
   ( (attribute->value (name-value 'foo "bar")) ==> "bar" )

   ( (attribute->name (attribute 'foobar "")) ==> foobar )
   ( (attribute->value (attribute 'foobar "")) ==> "" )
   ( (attribute->name (attribute 'foo "bar")) ==> foo )
   ( (attribute->value (attribute 'foo "bar")) ==> "bar" )

   ( (name-only? (attribute 'foobar "")) ==> #t )
   ( (name-value? (attribute 'foo "bar")) ==> #t )

   ( (child->element (child (name-value 'temp "32")))
     ==> (temp "32") )
   ( (text->string (text "a val")) ==> "a val" )
   ( (child-comment->comment (child-comment (comment "a comment")))
     ==> (*COMMENT* "a comment") )
   ( (child-pi->PI (PI 'targ "instr")) ==> (*PI* targ "instr") )
   ( (entity->public-id (entity "abc" "abc-1234")) ==> "abc" )
   ( (entity->system-id (entity "abc" "abc-1234")) ==> "abc-1234" )

    )
 )
