;; $Id: vector-product-mod.scm,v 1.2 2006/01/05 22:24:09 leavens Exp $

(module vector-product-mod (lib "typedscm.ss" "typedscm")

(provide vector-product)

;;; Examples:
;;; (vector-product '#(1 2 3 4)) ==> 24
;;; (vector-product '#(2 3 2 4)) ==> 48
;;; (vector-product '#(2 3 2 4 0 888 999)) ==> 0
;;; (vector-product '#()) ==> 1

(deftype vector-product (-> ((vector-of number)) number))
(define vector-product
  (lambda (von)
    ;; ENSURES: Result is the product of the elements in von
    (vector-product-iter von (- (vector-length von) 1) 1)))

;;; Design:                   von       n acc
;;;    (vector-product-iter '#(1 2 3 4) 3 1)
;;;  = (vector-product-iter '#(1 2 3 4) 2 4)
;;;  = (vector-product-iter '#(1 2 3 4) 1 12)
;;;  = (vector-product-iter '#(1 2 3 4) 0 24)
;;;  = (vector-product-iter '#(1 2 3 4) -1 24)
;;;  = 24
;;;
;;;    (vector-product-iter '#(1 3 0 4) 3 1)
;;;  = (vector-product-iter '#(1 3 0 4) 2 4)
;;;  = 0

(deftype vector-product-iter (-> ((vector-of number) number number) number))
(define vector-product-iter
  (lambda (von n acc)
    ;; REQUIRES: (< n (vector-length n))
    ;; ENSURES: Result is the product of acc and the elements in von whose
    ;;          index is less than or equal to n.
    (if (negative? n)
        acc
        (let ((elem (vector-ref von n)))
          (if (zero? elem)
              0
              (vector-product-iter
               von
               (- n 1)
               (* elem acc)))))))

) ;; end module
