;;; $Id: tc-logical-variables.scm,v 1.14 2006/02/02 04:19:56 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Logical variables used for unification and substition
;;; and procedures to ease creation/use.
;;;
;;; AUTHOR: Brian Dorn and Gary T. Leavens

(module tc-logical-variables (lib "typedscm.ss" "typedscm")

(provide tc:logicalvar? tc:variable? tc:make-logicalvar
         tc:symbol->logicalvar tc:new-logicalvar tc:new-lv-list
         tc:logicalvar->name tc:logicalvar->level tc:variable-eq?
	 tc:logicalvar->uniq-sym)

(deftype tc:logicalvar? (type-predicate-for tc:logicalvar))
(deftype tc:variable? (type-predicate-for symbol))
(deftype tc:make-logicalvar (-> (symbol number) tc:logicalvar))
(deftype tc:symbol->logicalvar (-> (symbol) tc:logicalvar))
(deftype tc:new-logicalvar (-> () tc:logicalvar))
(deftype tc:new-lv-list (-> (number) (list-of tc:logicalvar)))
(deftype tc:new-lv-list-maker
  (-> (number (list-of tc:logicalvar)) (list-of tc:logicalvar)))
(deftype tc:logicalvar->name (-> (tc:logicalvar) symbol))
(deftype tc:logicalvar->level (-> (tc:logicalvar) number))
(deftype tc:variable-eq? (-> (tc:logicalvar tc:logicalvar) boolean))
(deftype tc:logicalvar->uniq-sym (-> (tc:logicalvar) symbol))

(defrep (tc:logicalvar (pair-of symbol number)))

(define tc:logicalvar?
  (pair-of symbol? number?))

;; test a symbol to see if it looks like a variable
(define tc:variable?
  (has-type-trusted
   (type-predicate-for symbol)
   (lambda (x)
     (and (symbol? x)
	  (char=?
	   (string-ref (symbol->string x) 0) #\?)))))

(define tc:make-logicalvar
  (lambda (sym num)
    (cons sym num)))

(define tc:symbol->logicalvar
  (lambda (sym)
    (tc:make-logicalvar sym 0)))

;; tc:new-logicalvar is used to generate a fresh logicalvar
(define tc:new-logicalvar
  (let ((variable-count 0))
    (lambda ()
      (set! variable-count (+ variable-count 1))
      ;; (displayln "VAR NUM: " variable-count)
      (tc:symbol->logicalvar
       (string->symbol (string-append "?t" (number->string variable-count)))))))

(define tc:new-lv-list
  (lambda (n)
    (tc:new-lv-list-maker n '())))

(define tc:new-lv-list-maker
  (lambda (n l)
    (if (eqv? n 0)
	l
	(cons (tc:new-logicalvar)
	      (tc:new-lv-list-maker (- n 1) l)))))

(define tc:logicalvar->name
  (lambda (lv)
    (car lv)))

(define tc:logicalvar->level
  (lambda (lv)
    (cdr lv)))

(define tc:variable-eq? 
  (lambda (lv1 lv2)
    (eq? lv1 lv2)))

(define tc:logicalvar->uniq-sym
  (lambda (lvar)
    (string->symbol
     (string-append (symbol->string
		     (tc:logicalvar->name lvar))
		    "#"
		    (number->string (tc:logicalvar->level lvar))))))

) ;; end module
