;; (require (lib "defrep-with-define-datatype.scm" "typedscm" "tests" "bugs"))

(module defrep-with-define-datatype (lib "typedscm.ss" "typedscm")
  (provide texp texp? basic it ; it->b
	   binding? bind-texp bound? binding->texp binding->binder)

  (require (lib "relax-type-checking.scm" "typedscm"))

  (defrep (binding (pair-of symbol texp)))

  ;; The following gives an error:
  ;;    ERROR: tc:deep-down-type: invalid type expr: ...
  (deftype texp
    (variant-record
     (basic (name symbol))
     (it (b (pair-of symbol binding)))))
  (define-datatype texp texp?
    (basic (name symbol?))
    (it (b (pair-of symbol? binding?))))

  ;; The following gives an error in its inferred type vs. that declared.
  ;; This must be a problem with not using the defreps with define-datatype.
  (deftype it->b (-> (texp) binding))
  (define it->b
    (lambda (te)
      (cases texp te
	(it (b) b)
	(else (error "not an it" te)))))

  (deftype binding? (type-predicate-for binding))
  (define binding?
    (lambda (d)
      (and (pair? d)
	   (or (eq? 'let (car d))
	       (eq? 'letrec (car d))
	       (eq? 'lambda (car d)))
	   (texp? (cdr d)))))
  
  (deftype bind-texp (-> (symbol texp) binding))
  (define bind-texp
    (lambda (s te)
      (make-pair (has-type symbol s) (has-type texp te))))

  (deftype bound? (-> (symbol binding) boolean))
  (define bound?
    (lambda (s b)
      (eq? (pair-first (has-type (pair-of symbol texp) b)) s)))

  (deftype binding->texp (-> (binding) texp))
  (define binding->texp
    (lambda (b)
      (pair-second b)))

  (deftype binding->binder (-> (binding) symbol))
  (define binding->binder
    (lambda (b) (pair-first b)))

  ) ;; end module

;;; $Id: defrep-with-define-datatype.scm,v 1.1 2006/02/15 14:51:51 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.
